<?php
/**
 * Wicked storage implementation for PHP's PEAR database abstraction
 * layer.
 *
 * Required values for $params:<pre>
 *      'phptype'       The database type (e.g. 'pgsql', 'mysql', etc.).
 *      'hostspec'      The hostname of the database server.
 *      'protocol'      The communication protocol ('tcp', 'unix', etc.).
 *      'username'      The username with which to connect to the database.
 *      'password'      The password associated with 'username'.
 *      'database'      The name of the database.
 *      'table'         The name of the tasks table in 'database'.
 *      'charset'       The database's internal charset.</pre>
 *
 * Required by some database implementations:<pre>
 *      'options'       Additional options to pass to the database.
 *      'tty'           The TTY on which to connect to the database.
 *      'port'          The port on which to connect to the database.</pre>
 *
 * The table structure can be created by the scripts/drivers/wicked_foo.sql
 * script.
 *
 * $Horde: wicked/lib/Driver/sql.php,v 1.41 2004/11/18 14:39:40 eraserhd Exp $
 *
 * @author  Tyler Colbert <tyler-hordeml@colberts.us>
 * @version $Revision: 1.41 $
 * @since   Wicked 0.1
 * @package Wicked
 */
class Wicked_Driver_sql extends Wicked_Driver {

    /**
     * Handle for the current database connection.
     * @var object DB $_db
     */
    var $_db;

    /**
     * Boolean indicating whether or not we're connected to the SQL
     * server.
     * @var boolean $_connected
     */
    var $_connected = false;

    /**
     * Constructs a new Wicked SQL driver object.
     *
     * @param array  $params    A hash containing connection parameters.
     */
    function Wicked_Driver_sql($params = array())
    {
        parent::Wicked_Driver($params);
        $this->_connect();
    }

    /**
     * Retrieves the page of a particular name from the database.
     *
     * @param string  $pagename    The name of the page to retrieve.
     * @param boolean $autocreate  (optional) Try to create the page
     *                             from a standard page if we can't load it
     *                             from the database?. Defaults to false.
     *
     * @return mixed             True on success, PEAR_Error on failure.
     */
    function retrieveByName($pagename, $autocreate = false, $template = null)
    {
        $where = sprintf('page_name = %s',
                         $this->_db->quote($pagename));

        $pages = $this->_retrieve($this->_params['table'], $where);
        if (is_a($pages, 'PEAR_Error')) {
            Horde::logMessage($pages, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $pages;
        }

        if (!empty($pages[0])) {
            return $pages[0];
        }

        if (!$autocreate) {
            /* At this point we have to give up, since we found no
             * data and we're not allowed to try creating the page. */
            return PEAR::raiseError($pagename . ' not found');
        }

        /* If we didn't have that page in the database, try and create
         * it, either from the static initial pages in scripts/data/
         * or from scratch. */
        return $this->loadPage($pagename, $template);
    }

    function retrieveHistory($pagename, $version)
    {
        if (empty($version) or !preg_match('/^[0-9]+\.[0-9]+$/',$version)) {
            return PEAR::raiseError('invalid version number');
        }

        list($major, $minor) = explode('.',$version);
        $where = sprintf('page_name = %s AND page_majorversion = %s AND ' .
                         'page_minorversion = %s',
                         $this->_db->quote($pagename),
                         $major, $minor);

        return $this->_retrieve($this->_params['historytable'], $where);
    }

    function getPage($pagename)
    {
        $where = sprintf('page_name = %s', $this->_db->quote($pagename));
        return $this->_retrieve($this->_params['table'], $where);
    }

    function getAllPages()
    {
        return $this->_retrieve($this->_params['table'], '', 'page_name');
    }

    function getHistory($pagename)
    {
        $where = sprintf('page_name = %s ORDER BY page_majorversion DESC, page_minorversion DESC',
                         $this->_db->quote($pagename));

        return $this->_retrieve($this->_params['historytable'], $where);
    }

    function getRecentChanges()
    {
        $where = sprintf('version_created > %s',
                         time() - (86400 * 3));

        $result = $this->_retrieve($this->_params['table'], $where, 'version_created DESC');
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        $result2 = $this->_retrieve($this->_params['historytable'], $where, 'version_created DESC');
        if (is_a($result2, 'PEAR_Error')) {
            return $result2;
        }

        return array_merge($result, $result2);
    }

    function mostPopular($limit = 10)
    {
        return $this->_retrieve($this->_params['table'], '', 'page_hits DESC', $limit);
    }

    function leastPopular($limit = 10)
    {
        return $this->_retrieve($this->_params['table'], '', 'page_hits ASC', $limit);
    }

    function searchTitles($searchtext)
    {
        require_once 'Horde/SQL.php';
        $where = Horde_SQL::buildClause($this->_db, 'page_name', 'LIKE', $searchtext);

        return $this->_retrieve($this->_params['table'], $where);
    }

    function searchText($searchtext)
    {
        require_once 'Horde/SQL/Keywords.php';

        $nameClause = Horde_SQL_Keywords::parse("page_name", $searchtext);
        if (is_a($nameClause, 'PEAR_Error')) {
            return $nameClause;
        }

        $textClause = Horde_SQL_Keywords::parse("page_text", $searchtext);
        if (is_a($textClause, 'PEAR_Error')) {
            return $textClause;
        }

        $where = sprintf("( %s ) OR ( %s )", $nameClause, $textClause);
        return $this->_retrieve($this->_params['table'], $where);
    }

    function getBackLinks($pagename)
    {
        $where = sprintf("page_text LIKE %s",
                         $this->_db->quote('%' . $pagename . '%'));
        $pages = $this->_retrieve($this->_params['table'], $where);
        if (is_a($pages, 'PEAR_Error')) {
            return $pages;
        }

        /* We've cast a wide net, so now we filter out pages which don't
         * actually refer to $pagename. */

        $patterns = array('/\(\(' . preg_quote($pagename, '/') . '\)\)/');
        if (preg_match('/^' . WICKED_REGEXP_WIKIWORD . '$/', $pagename)) {
            $patterns[] = '/\b' . preg_quote($pagename, '/') . '\b/';
        }

        foreach ($pages as $key => $page) {
            $match = false;
            foreach ($patterns as $pattern) {
                if (preg_match($pattern, $page['page_text'])) {
                    $match = true;
                }
            }
            if (!$match) {
                unset($pages[$key]);
            }
        }

        return $pages;
    }

    function getLikePages($pagename)
    {
        if (ctype_upper($pagename)) {
            $firstword = $pagename;
            $lastword = null;
        } else {
            /* Get the first and last word of the page name. */
            $count = preg_match_all('/[A-Z][a-z]*/', $pagename, $matches);
            $matches = $matches[0];

            $firstword = $matches[0];
            $lastword = $matches[$count - 1];

            if (strlen($firstword) == 1 && strlen($matches[1]) == 1) {
                for ($i = 1; $i < $count; $i++) {
                    $firstword .= $matches[$i];
                    if (isset($matches[$i + 1]) && strlen($matches[$i + 1]) > 1) {
                        break;
                    }
                }
            }

            if (strlen($lastword) == 1 && strlen($matches[$count - 2]) == 1) {
                for ($i = $count - 2; $i > 0; $i--) {
                    $lastword = $matches[$i] . $lastword;
                    if (isset($matches[$i - 1]) && strlen($matches[$i - 1]) > 1) {
                        break;
                    }
                }
            }
        }

        require_once 'Horde/SQL.php';

        $where = Horde_SQL::buildClause($this->_db, 'page_name', 'LIKE', $firstword);
        if (!empty($lastword) && $lastword != $firstword) {
            $where .= ' OR ' . Horde_SQL::buildClause($this->_db, 'page_name', 'LIKE', $lastword);
        }

        return $this->_retrieve($this->_params['table'], $where);
    }

    function getAttachedFiles($pagename, $allversions = false)
    {
        $where = sprintf('page_name = %s', $this->_db->quote($pagename));
        $data = $this->_retrieve($this->_params['attachmenttable'], $where);
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }

        if ($allversions) {
            $more_data = $this->_retrieve($this->_params['attachmenthistorytable'], $where);
            if (is_a($more_data, 'PEAR_Error')) {
                return $more_data;
            }
            $data = array_merge($data, $more_data);
        }

        usort($data, create_function('$a, $b','
            $res = strcmp($a["attachment_name"], $b["attachment_name"]);
            if ($res != 0) {
                return $res;
            }
            $res = ($a["attachment_majorversion"] - $b["attachment_minorversion"]);
            if ($res != 0) {
                return $res;
            }

            return ($a["attachment_minorversion"] - $b["attachment_minorversion"]);
        '));
        return $data;
    }

    function _attachFile($file)
    {
        if (is_null($file['change_author'])) {
            $file['change_author'] = Auth::getAuth();
        }

        $where = sprintf('page_name = %s AND attachment_name = %s',
                         $this->_db->quote($file['page_name']),
                         $this->_db->quote($file['attachment_name']));
        $attachments = $this->_retrieve($this->_params['attachmenttable'],
                                        $where);
        if (is_a($attachments, 'PEAR_Error')) {
            return $attachments;
        }

        if (count($attachments) > 0) {
            list($old) = $attachments;
            $majorversion = $old['attachment_majorversion'];
            $minorversion = $old['attachment_minorversion'];
            if ($file['minor']) {
                $minorversion++;
            } else {
                $majorversion++;
                $minorversion = 0;
            }

            $sql = sprintf("INSERT INTO %s (page_name, attachment_name, attachment_majorversion, attachment_minorversion, attachment_created, change_author, change_log) SELECT page_name, attachment_name, attachment_majorversion, attachment_minorversion, attachment_created, change_author, change_log FROM %s WHERE page_name = %s AND attachment_name = %s;",
                           $this->_params['attachmenthistorytable'],
                           $this->_params['attachmenttable'],
                           $this->_db->quote($file['page_name']),
                           $this->_db->quote($file['attachment_name']));
            $res = $this->_db->query($sql);
            if (is_a($res, 'PEAR_Error')) {
                var_dump($res);
                return $res;
            }

            $sql = sprintf("UPDATE %s SET attachment_majorversion = %s, attachment_minorversion = %s, change_log = %s, change_author = %s, attachment_created = %s WHERE page_name = %s AND attachment_name = %s;",
                           $this->_params['attachmenttable'],
                           intval($majorversion),
                           intval($minorversion),
                           $this->_db->quote($file['change_log']),
                           $this->_db->quote($file['change_author']),
                           intval(time()),
                           $this->_db->quote($file['page_name']),
                           $this->_db->quote($file['attachment_name']));
            $res = $this->_db->query($sql);
            if (is_a($res, 'PEAR_Error')) {
                return $res;
            }

        } else {
            $majorversion = 1;
            $minorversion = 0;
            $sql = sprintf("INSERT INTO %s (attachment_majorversion, attachment_minorversion, change_log, change_author, attachment_created, page_name, attachment_name) VALUES ( 1, 0, %s, %s, %s, %s, %s );",
                           $this->_params['attachmenttable'],
                           $this->_db->quote($file['change_log']),
                           $this->_db->quote($file['change_author']),
                           intval(time()),
                           $this->_db->quote($file['page_name']),
                           $this->_db->quote($file['attachment_name']));
            $res = $this->_db->query($sql);
            if (is_a($res, 'PEAR_Error')) {
                return $res;
            }
        }

        return sprintf("%d.%d", $majorversion, $minorversion);
    }

    /**
     * Log a hit to $pagename.
     *
     * @param string $pagename  The page that was viewed.
     *
     * @return mixed  True or PEAR_Error on failure.
     */
    function logPageView($pagename)
    {
        $query = sprintf('UPDATE %s SET page_hits = page_hits + 1 WHERE page_name = %s',
                         $this->_params['table'],
                         $this->_db->quote($pagename));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::logPageView(%s): %s', $pagename, $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Execute the query. */
        return $this->_db->query($query);
    }

    function lock($pagename, $locked)
    {
        if ($locked) {
            $query = sprintf('UPDATE %s SET page_locked = 1 WHERE page_name = %s',
                             $this->_params['table'],
                             $this->_db->quote($pagename));
        } else {
            $query = sprintf('UPDATE %s SET page_locked = 0 WHERE page_name = %s',
                             $this->_params['table'],
                             $this->_db->quote($pagename));
        }

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::lock(%s, %s): %s', $pagename, $locked, $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Execute the query. */
        return $this->_db->query($query);
    }

    function newPage($pagename, $text)
    {
        $author = Auth::getAuth();
        if ($author === false) {
            $author = null;
        }

        $query = sprintf(
            'INSERT INTO %s (page_name, page_text, ' .
            'page_locked, version_created, ' .
            'page_majorversion, page_minorversion, page_hits, change_author) ' .
            'VALUES (%s, %s, 0, %s, 1, 0, 0, %s)',
            $this->_params['table'],
            $this->_db->quote($pagename),
            $this->_db->quote(String::convertCharset($text, NLS::getCharset(), $this->getCharset())),
            time(),
            $this->_db->quote($author));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::newPage(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the insertion/update query. */
        $result = $this->_db->query($query);

        /* Return an error immediately if the query failed. */
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return true;
    }

    /**
     * Rename a page (and keep the page's history).
     *
     * @param string $pagename          the name of the page to rename.
     * @param string $newname           the page's new name.
     *
     * @return mixed  True or PEAR_Error on failure.
     */
    function renamePage($pagename, $newname)
    {
        $query = sprintf("UPDATE %s SET page_name = %s WHERE page_name = %s;",
                         $this->_params['table'],
                         $this->_db->quote($newname),
                         $this->_db->quote($pagename));

        Horde::logMessage(sprintf('Wicked_Driver_sql::renamePage(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        $query = sprintf("UPDATE %s SET page_name = %s WHERE page_name = %s;",
                         $this->_params['historytable'],
                         $this->_db->quote($newname),
                         $this->_db->quote($pagename));

        Horde::logMessage(sprintf('Wicked_Driver_sql::renamePage(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        $changelog = sprintf(_("Renamed page from %s"), $pagename);
        $newPage = $this->retrieveByName($newname);
        if (is_a($newPage, 'PEAR_Error')) {
            return $newPage;
        }

        return $this->updateText($newname, $newPage['page_text'], $changelog,
                                 true);
    }

    function updateText($pagename, $text, $changelog, $minorchange)
    {
        if (!$this->pageExists($pagename)) {
            return $this->newPage($pagename, $text);
        }

        /* Copy the old version into the page history. */
        $query = sprintf(
            'INSERT INTO %s (page_name, page_text, page_majorversion, page_minorversion, version_created, change_author, change_log)' .
            ' SELECT page_name, page_text, page_majorversion, page_minorversion, version_created, change_author, change_log FROM %s WHERE page_name = %s',
            $this->_params['historytable'],
            $this->_params['table'],
            $this->_db->quote($pagename));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::updateText(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);

        /* Return an error immediately if the query failed. */
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        /* Now move on to updating the record. */
        if ($minorchange) {
            $versionchange = 'page_minorversion = page_minorversion + 1 ';
        } else {
            $versionchange = 'page_majorversion = page_majorversion + 1, page_minorversion = 0 ';
        }

        $author = Auth::getAuth();
        if ($author === false) {
            $author = null;
        }

        $query = sprintf(
            'UPDATE %s SET change_author = %s, page_text = %s, change_log = %s, version_created = %s, ' . $versionchange . 'WHERE page_name = %s',
            $this->_params['table'],
            $this->_db->quote($author),
            $this->_db->quote(String::convertCharset($text, NLS::getCharset(), $this->getCharset())),
            $this->_db->quote(String::convertCharset($changelog, NLS::getCharset(), $this->getCharset())),
            time(),
            $this->_db->quote($pagename));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::updateText(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the insertion/update query. */
        $result = $this->_db->query($query);

        /* Return an error immediately if the query failed. */
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        return true;
    }

    function getPages()
    {
        $query = sprintf('SELECT page_name FROM %s ',
                         $this->_params['table']);

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::getPages(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        $pageNames = array();
        while ($row = $result->fetchRow()) {
            $pageNames[] = $row[0];
        }

        return $pageNames;
    }

    /**
     * @TODO If the top version is deleted, move another one up.
     */
    function removeVersion($pagename, $version)
    {
        list($major, $minor) = explode('.', $version);

        $query = sprintf(
            'DELETE FROM %s ' .
            'WHERE page_name = %s AND page_majorversion = %d AND page_minorversion = %d',
            $this->_params['table'],
            $this->_db->quote($pagename),
            $major, $minor);

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::removeVersion(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);

        $query = sprintf(
            'DELETE FROM %s ' .
            'WHERE page_name=%s and page_majorversion = %d and page_minorversion = %d',
            $this->_params['historytable'],
            $this->_db->quote($pagename),
            $major, $minor);

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::removeVersion(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);
    }

    function removeAllVersions($pagename)
    {
        $this->_pageNames = null;

        $query = sprintf('DELETE FROM %s WHERE page_name = %s',
                         $this->_params['table'],
                         $this->_db->quote($pagename));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::removeAllVersions(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        $query = sprintf('DELETE FROM %s WHERE page_name = %s',
                         $this->_params['historytable'],
                         $this->_db->quote($pagename));

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::removeAllVersions(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Attempt the query. */
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
            return $result;
        }

        /* Send notification. */
        return parent::removeAllVersions($pagename);
    }

    /**
     * Retrieves a page or set of pages given an SQL WHERE clause.
     *
     * @param string $table     Which table are we retrieving pages from?
     * @param string $sqlWhere  Where clause for sql statement (without the 'WHERE').
     * @param string $orderBy   What column should we order results by?
     *
     * @return array | object  Either an array of pages or PEAR::Error.
     */
    function _retrieve($table, $sqlWhere, $orderBy = null, $limit = null)
    {
        $query = sprintf('SELECT * FROM %s%s%s',
                         $table,
                         !empty($sqlWhere) ? ' WHERE ' . $sqlWhere : '',
                         !empty($orderBy) ? ' ORDER BY ' . $orderBy : '');

        if (!empty($limit)) {
            $query = $this->_db->modifyLimitQuery($query, 0, $limit);
        }

        /* Log the query at a DEBUG log level. */
        Horde::logMessage(sprintf('Wicked_Driver_sql::_retrieve(): %s', $query),
                          __FILE__, __LINE__, PEAR_LOG_DEBUG);

        /* Execute the query. */
        $result = $this->_db->query($query);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        $row = $result->fetchRow(DB_FETCHMODE_ASSOC);
        if (is_a($row, 'PEAR_Error')) {
            return $row;
        }

        $pages = array();
        $index = 0;
        while ($row && !is_a($row, 'PEAR_Error')) {
            $pages[$index] = $row;
            $pages[$index]['locked'] = !empty($row['page_locked']);

            /* Advance to the new row in the result set. */
            $row = $result->fetchRow(DB_FETCHMODE_ASSOC);
            $index++;
        }
        $result->free();

        return $pages;
    }

    /**
     * Returns the charset used by the backend.
     *
     * @return string  The backend's charset
     */
    function getCharset()
    {
        return $this->_params['charset'];
    }

    /**
     * Attempts to open a persistent connection to the SQL server.
     *
     * @return boolean    True on success; exits (Horde::fatal()) on error.
     */
    function _connect()
    {
        if (!$this->_connected) {
            Horde::assertDriverConfig($this->_params, 'storage',
                                      array('phptype', 'hostspec', 'username', 'database', 'charset'));

            if (!isset($this->_params['table'])) {
                $this->_params['table'] = 'wicked_pages';
            }
            if (!isset($this->_params['historytable'])) {
                $this->_params['historytable'] = 'wicked_history';
            }
            if (!isset($this->_params['attachmenttable'])) {
                $this->_params['attachmenttable'] = 'wicked_attachments';
            }
            if (!isset($this->_params['attachmenthistorytable'])) {
                $this->_params['attachmenthistorytable'] = 'wicked_attachment_history';
            }

            /* Connect to the SQL server using the supplied
             * parameters. */
            require_once 'DB.php';
            $this->_db = &DB::connect($this->_params,
                                      array('persistent' => !empty($this->_params['persistent'])));
            if (is_a($this->_db, 'PEAR_Error')) {
                Horde::fatal($this->_db, __FILE__, __LINE__);
            }

            $this->_db->setOption('portability', DB_PORTABILITY_LOWERCASE | DB_PORTABILITY_ERRORS);

            $this->_connected = true;
        }

        return true;
    }

}
